# SafeSpawn by David Bowland
# ./addons/eventscripts/safespawn/safespawn.py

# >>> To configure this addon please see safespawn.cfg (created when the addon is first loaded) <<<

"""
Provides protection from damage after spawn. Requires only ES 2.1
"""

from __future__ import division

import cfglib
import es
import gamethread
import langlib
import playerlib

import psyco
psyco.full()

from path import path
from time import time as now
from usermsg import hudhint


info = es.AddonInfo()
info.name     = 'SafeSpawn'
info.version  = '3'
info.basename = 'safespawn'
info.url      = 'http://addons.eventscripts.com/addons/view/' + info.basename
info.author   = 'SuperDave'


addonpath = path(__file__).dirname()

###

config = cfglib.AddonCFG(addonpath.joinpath(info.basename + '.cfg'))

config.text(info.name + ' release %s options' % info.version)
config.text('./addons/eventscripts/%(basename)s/%(basename)s.cfg' % {'basename': info.basename})
config.text(info.url)
config.text('')
config.text('Load this addon with: es_load ' + info.basename)
config.text('\n')

config.text('***** Protection options *****')
cvar_protection_time     = config.cvar('safespawn_protection_time',     6, 'Number of seconds after spawning players will be invincible (should be greater than or equal to 1)')
cvar_protection_perspawn = config.cvar('safespawn_protection_perspawn', 2, '0 = spawn protection starts when the round starts, 1 = spawn protection starts when a player spawns (deathmatch-style), 2 = both round start and player spawn protection')
config.text('\n')

config.text('***** Fade options *****')
cvar_fade_steps = config.cvar('safespawn_fade_steps', 4,   'Number of fading steps before players become fully visible, 0 = disabled')
cvar_fade_alpha = config.cvar('safespawn_fade_alpha', 100, 'Alpha value to give players when they become invincible (0, invisible to 255, fully visible)')
config.text('\n')

config.text('***** Announcement options *****')
cvar_announce_on_chat = config.cvar('safespawn_announce_on_chat', 0, '1 = announce in the chat area when spawn protection begins, 0 = disabled')
cvar_announce_on_center = config.cvar('safespawn_announce_on_center', 0, '1 = announce in center text when spawn protection begins, 0 = disabled')
cvar_announce_on_hud = config.cvar('safespawn_announce_on_hud', 0, '1 = announce in a HUD hint when spawn protection begins, 0 = disabled')
config.text('\n')
cvar_announce_off_chat = config.cvar('safespawn_announce_off_chat', 0, '1 = announce in the chat area when spawn protection ends, 0 = disabled')
cvar_announce_off_center = config.cvar('safespawn_announce_off_center', 1, '1 = announce in center text when spawn protection ends, 0 = disabled')
cvar_announce_off_hud = config.cvar('safespawn_announce_off_hud', 0, '1 = announce in a HUD hint when spawn protection ends, 0 = disabled')

config.write()

config.execute()

###

ini = cfglib.AddonINI(addonpath.joinpath(info.basename + '_languages.ini'))

ini.setInitialComments([
 '# ./addons/eventscripts/%(basename)s/%(basename)s_languages.ini' % {'basename': info.basename},
 '# >>> For general addon configuration see ' + info.basename + '.cfg <<<',
 '',
 '# DO NOT translate words with $ or # in front!',
 '# For a list of language abbreviations please see ./addons/eventscripts/_libs/python/deflangs.ini'])

ini.addGroup('protection on')
ini.addValueToGroup('protection on', 'en', '#defaultSpawn protection: #darkgreenOn')

ini.addGroup('protection off')
ini.addValueToGroup('protection off', 'en', '#defaultSpawn protection: #darkgreenOff')

ini.write()
lang_text = langlib.Strings(ini)


def removeTags(text):
   """ Returns text without color tags """
   return reduce(lambda t, r: t.replace(r, ''), ('#lightgreen', '#darkgreen', '#green', '#default'), text)

###

class BasePlayer(object):
   def removeProtection(self):
      pass

#

class SafeSpawnPlayer(BasePlayer):
   alpha = 255

   def __init__(self, userid):
      self.userid = userid
      self.delayname = 'safespawn_' + str(userid)

   def startProtection(self, duration=cvar_protection_time):
      self.removeDelay()

      if cvar_fade_steps:
         alpha = self.alpha = int(cvar_fade_alpha)
         if alpha != 255:
            prot_time = float(cvar_protection_time)
            fade_steps = int(cvar_fade_steps)
            # If we have a delayed spawn we might want to start 
            startstep = min(fade_steps - 1, # Can't start higher than the last step
             int((prot_time - duration) * fade_steps / prot_time))
            self.executeStep(startstep, fade_steps, now() + duration, prot_time / fade_steps)

      # Make player invincible and start delay for protection to expire
      self.setHitbox(False)
      gamethread.delayedname(duration, self.delayname, self.removeProtection)

      # Make necessary announcements
      if any((cvar_announce_on_chat, cvar_announce_on_center, cvar_announce_on_hud)):
         userid = self.userid
         text = lang_text('protection on', lang=playerlib.getPlayer(userid).getLanguage())
         if cvar_announce_on_chat:
            es.tell(userid, '#multi', text)
         if cvar_announce_on_center:
            es.centertell(userid, removeTags(text))
         if cvar_announce_on_hud:
            hudhint(userid, removeTags(text))

   def executeStep(self, step, total_steps, endtime, step_duration):
      self.setAlpha(cvar_fade_alpha + (255 - cvar_fade_alpha) * step / total_steps)
      step += 1
      if step < total_steps:
         # This is done to sync "steps" between players with the same end time but different start times
         gamethread.delayedname((endtime - step_duration * (total_steps - step)) - now(),
          self.delayname, self.executeStep, (step, total_steps, endtime, step_duration))

   def removeProtection(self):
      self.removeDelay()
      self.setHitbox(True)
      if self.alpha != 255:
         self.setAlpha(255)

      # Make necessary announcements
      if any((cvar_announce_off_chat, cvar_announce_off_center, cvar_announce_off_hud)):
         userid = self.userid
         text = lang_text('protection off', lang=playerlib.getPlayer(userid).getLanguage())
         if cvar_announce_off_chat:
            es.tell(userid, '#multi', text)
         if cvar_announce_off_center:
            es.centertell(userid, removeTags(text))
         if cvar_announce_off_hud:
            hudhint(userid, removeTags(text))

      del ss[self.userid]

   def removeDelay(self):
      gamethread.cancelDelayed(self.delayname)

   def setAlpha(self, value):
      player = playerlib.getPlayer(self.userid)
      player.setColor(*(player.getColor()[:3] + (int(value),)))

   def setHitbox(self, value):
      es.setplayerprop(self.userid, 'CBaseAnimating.m_nHitboxSet', 0 if value else 2)

###

class SafeSpawnManager(dict):
   def __getitem__(self, userid):
      userid = int(userid)
      if userid in self:
         return super(SafeSpawnManager, self).__getitem__(userid)
      x = self[userid] = SafeSpawnPlayer(userid)
      return x

   def __delitem__(self, userid):
      userid = int(userid)
      if userid in self:
         self[userid].removeDelay()
         super(SafeSpawnManager, self).__delitem__(userid)

   def clear(self):
      for player in self.itervalues():
         player.removeDelay()
      super(SafeSpawnManager, self).clear()

ss = SafeSpawnManager()

###

class RoundTimeManager(object):
   starttime = 0

   def roundStart(self):
      self.starttime = now()

   def getProtectionTime(self):
      return (self.starttime + cvar_protection_time) - now()

roundtime = RoundTimeManager()

###

def es_map_start(event_var):
   ss.clear()


def player_spawn(event_var):
   if cvar_protection_perspawn == 1:
      ss[event_var['userid']].startProtection()
   else:
      protectiontime = roundtime.getProtectionTime()
      if protectiontime > 0:
         ss[event_var['userid']].startProtection(protectiontime)
      elif cvar_protection_perspawn:
         ss[event_var['userid']].startProtection()


def round_start(event_var):
   roundtime.roundStart()
   if cvar_protection_perspawn != 1:
      for userid in es.getUseridList():
         if not es.getplayerprop(userid, 'CBasePlayer.pl.deadflag'):
            protectiontime = roundtime.getProtectionTime()
            if protectiontime > 0:
               ss[userid].startProtection(protectiontime)


def player_hurt(event_var):
   attacker = int(event_var['attacker'])
   if attacker in ss:
      ss[attacker].removeProtection()


def player_death(event_var):
   del ss[event_var['userid']]


def player_disconnect(event_var):
   del ss[event_var['userid']]


def unload():
   ss.clear()

   for cvar in config.getCvars():
      es.ServerVar(cvar).set(0)